using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Syntriax.Engine.Core.Abstract;
using Syntriax.Engine.Physics2D.Abstract;

namespace Syntriax.Engine.Physics2D;

public class PhysicsEngine2D : IPhysicsEngine2D
{
    private IList<IRigidBody2D> rigidBodies = new List<IRigidBody2D>(32);
    private IList<ICollider2D> colliders = new List<ICollider2D>(64);

    private int _iterationCount = 1;


    public int IterationCount { get => _iterationCount; set => _iterationCount = value < 1 ? 1 : value; }

    public void AddRigidBody(IRigidBody2D rigidBody)
    {
        if (rigidBodies.Contains(rigidBody))
            return;

        rigidBodies.Add(rigidBody);

        foreach (var collider2D in rigidBody.BehaviourController.GetBehaviours<ICollider2D>())
            colliders.Add(collider2D);

        rigidBody.BehaviourController.OnBehaviourAdded += OnBehaviourAdded;
        rigidBody.BehaviourController.OnBehaviourRemoved += OnBehaviourRemoved;
    }

    public void RemoveRigidBody(IRigidBody2D rigidBody)
    {
        rigidBodies.Remove(rigidBody);
    }

    public void Step(float deltaTime)
    {
        float intervalDeltaTime = deltaTime / IterationCount;

        for (int i = 0; i < IterationCount; i++)
        {
            foreach (var rigidBody in rigidBodies)
            {
                Vector2 nextPosition = rigidBody.Transform.Position;
                nextPosition += rigidBody.Velocity * intervalDeltaTime;

                rigidBody.Transform.Position = nextPosition;
            }


            foreach (var collider in colliders)
                collider.RecalculateVertices();

            for (int colliderIX = 0; colliderIX < colliders.Count; colliderIX++)
            {
                ICollider2D colliderX = colliders[colliderIX];
                for (int colliderIY = colliderIX + 1; colliderIY < colliders.Count; colliderIY++)
                    foreach (var vertex in colliderX.Vertices)
                    {
                        if (!colliders[colliderIY].CheckCollision(vertex, colliderX))
                            continue;

                        Console.WriteLine($"Collision");
                        break;
                    }
            }
        }
    }


    private void OnBehaviourAdded(IBehaviourController controller, IBehaviour behaviour)
    {
        if (behaviour is not ICollider2D collider2D)
            return;

        colliders.Add(collider2D);
    }

    private void OnBehaviourRemoved(IBehaviourController controller, IBehaviour behaviour)
    {
        if (behaviour is not ICollider2D collider2D)
            return;

        colliders.Remove(collider2D);
    }
}
